"""MIT License

Copyright (c) 2025 Omkaar

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


# pylint: skip-file

import unittest
import xkcd

class TestWhatIfArticle(unittest.TestCase):
    def test_latest_article(self):
        article = xkcd.WhatIfArticle()
        self.assertIsInstance(article.number, int)
        self.assertIsInstance(article.title, str)
        self.assertIsInstance(article.question, str)
        self.assertIsInstance(article.author, str)
        self.assertTrue(article.url.startswith("https://what-if.xkcd.com/"))
        self.assertIsInstance(article.entry, list)
        # Check entry types (str, Image, Hyperlink, Reference)
        for item in article.entry:
            self.assertTrue(
                isinstance(item, (str, xkcd.WhatIfArticle.Image, xkcd.WhatIfArticle.Hyperlink, xkcd.WhatIfArticle.Reference))
            )

    def test_random_article(self):
        article = xkcd.WhatIfArticle(random=True)
        self.assertIsInstance(article.number, int)

    def test_specific_article(self):
        article = xkcd.WhatIfArticle(1)
        self.assertEqual(article.number, 1)
        self.assertIsInstance(article.title, str)

    def test_image_class(self):
        article = xkcd.WhatIfArticle(1)
        images = [e for e in article.entry if isinstance(e, xkcd.WhatIfArticle.Image)]
        for img in images:
            self.assertIsInstance(img.url, str)
            self.assertIsInstance(img.title, str)
            self.assertIsInstance(img.filename, str)

    def test_hyperlink_class(self):
        article = xkcd.WhatIfArticle(1)
        links = [e for e in article.entry if isinstance(e, xkcd.WhatIfArticle.Hyperlink)]
        for link in links:
            self.assertIsInstance(link.text, str)
            self.assertIsInstance(link.url, str)

    def test_reference_class(self):
        article = xkcd.WhatIfArticle(1)
        refs = [e for e in article.entry if isinstance(e, xkcd.WhatIfArticle.Reference)]
        for ref in refs:
            self.assertIsInstance(ref.number, int)
            self.assertIsInstance(ref.text, str)

    def test_invalid_random_and_number(self):
        with self.assertRaises(ValueError):
            xkcd.WhatIfArticle(1, random=True)

    def test_invalid_number(self):
        # Try a very large number, should raise ValueError
        with self.assertRaises(ValueError):
            xkcd.WhatIfArticle(999999)

if __name__ == "__main__":
    unittest.main()



"""
MIT License

Copyright (c) 2025 Omkaar

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from datetime import datetime
from html import unescape
from os.path import split
from random import randint
from urllib.parse import urlparse
from typing import Optional

from requests import get


XKCD_BASE_URL = "https://xkcd.com/"
XKCD_WIKI_BASE_URL = "https://explainxkcd.com/"


class Comic:

    """
    A class that represents a comic.

    :param number: The comic's number.
    :type number: Optional[:class:`int`]
    :param random: Whether to choose a random comic, or not.
    :type random: Optional[:class:`bool`]

    .. note::

        If ``random`` is ``True``, ``number`` must not be specified.

    :ivar date: The comic's date.
    :ivar image: The URL of the comic's image.
    :ivar number: The number of the comic.
    :ivar title: The comic's title.
    :ivar safe_title: A safe form of the comic's title.
    :ivar transcript: The trascript of the comic.
    :ivar wiki_url: The URL of the comic's wiki.
    :ivar url: The comic's URL.
    """

    class Image:

        """
        A class that represents an image.

        :ivar url: The image's URL.
        :ivar title: The image's title (Alt Text).
        :ivar filename: The filename of the image.
        """

        def __init__(self, _url: str, _title: str) -> None:
            self.url = _url
            self.title = _title
            self.filename = split(urlparse(self.url).path)[1]

    def __init__(self, number: Optional[int] = None, *, random: Optional[bool] = False) -> None:

        if random and number:
            raise ValueError("If 'random' is 'True', 'number' must not be specified.")

        if random:
            response = get(f"{XKCD_BASE_URL}info.0.json").json()
            latest = int(response["num"])
            self.number = randint(1, latest)
            response = get(f"{XKCD_BASE_URL}{self.number}/info.0.json").json()
        else:
            request_url = f"{XKCD_BASE_URL}info.0.json" if number is None else f"{XKCD_BASE_URL}{number}/info.0.json"
            response = get(request_url).json()
            self.number = int(response["num"])
        self.date = datetime(int(response["year"]), int(response["month"]), int(response["day"]))
        self.safe_title = response["safe_title"]
        self.title = response["title"]
        self.transcript = unescape(response["transcript"])
        self.image = self.Image(response["img"], response["alt"])

        self.wiki_url = f"{XKCD_WIKI_BASE_URL}{self.number}"
        self.url = f"{XKCD_BASE_URL}{self.number}"



"""
xkcd API Wrapper
~~~~~~~~~~~~~~~~~~~

A basic API wrapper for xkcd.

:copyright: (c) 2025 Omkaar
:license: MIT, see LICENSE for more details.
"""


__title__ = "xkcd.py"
__author__ = "Ombucha"
__license__ = "MIT"
__copyright__ = "Copyright 2025 Omkaar"
__version__ = "1.1.0"


from .comic import Comic
from .what_if import WhatIfArticle



"""
MIT License

Copyright (c) 2025 Omkaar

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from os.path import split
from random import randint
from urllib.parse import urlparse
from urllib.request import urlopen, Request
from typing import Optional

from bs4 import BeautifulSoup
from bs4.element import NavigableString, Tag


WHAT_IF_BASE_URL = "https://what-if.xkcd.com/"


class WhatIfArticle:

    """
    A class that represents a What If article.

    :param number: The article's number.
    :type number: Optional[:class:`int`]
    :param random: Whether to choose a random article, or not.
    :type random: Optional[:class:`bool`]

    .. note::

        If ``random`` is ``True``, ``number`` must not be specified.

    :ivar entry: The article entry. It is a list of :class:`str`, :class:`Image`, :class:`Hyperlink` and :class:`Reference`.
    :ivar number: The number of the article.
    :ivar title: The article's title.
    :ivar question: The question of the article.
    :ivar author: The author of the article.
    :ivar url: The article's URL.
    """

    class Image:

        """
        A class that represents an image.

        :ivar url: The image's URL.
        :ivar title: The image's title (Alt Text).
        :ivar filename: The filename of the image.
        """

        def __init__(self, _url: str, _title: str) -> None:
            self.url = _url
            self.title = _title
            self.filename = split(urlparse(self.url).path)[1]

    class Hyperlink:

        """
        A class that represents hyperlinked text.

        :ivar url: The URL that hyperlink leads to.
        :ivar text: The text shown on the hyperlink.
        """

        def __init__(self, _text: str, _url: str) -> None:
            self.text = _text
            self.url = _url

    class Reference:

        """
        A class that represents a reference.

        :ivar number: The reference's number.
        :ivar text: The text shown upon hovering over the reference.
        """

        def __init__(self, _number: int, _text: str) -> None:
            self.number = _number
            self.text = _text

    def __init__(self, number: Optional[int] = None, *, random = False) -> None:

        if random and number:
            raise ValueError("If 'random' is 'True', 'number' must not be specified.")

        page = Request(f"{WHAT_IF_BASE_URL}archive")
        with urlopen(page) as result:
            soup = BeautifulSoup(result.read(), "html.parser")

        entries = soup.find_all("div", {"class": "archive-entry"})
        latest = int(entries[-1].a.attrs["href"].split("/")[-1])
        if random:
            self.number = randint(1, latest)
        else:
            if number is None:
                self.number = latest
            else:
                if number > latest:
                    raise ValueError("You have chosen a comic after the latest one.")
                self.number = number

        page = Request(f"{WHAT_IF_BASE_URL}{self.number}")
        with urlopen(page) as result:
            soup = BeautifulSoup(result.read(), "html.parser")

        self.entry = []
        for tag in soup.find_all():
            identifier = not "id" in tag.attrs.keys()
            if tag.parent.name == "article" and identifier:
                children = list(tag.children)
                if tag.name == "p" and len(children) > 0:
                    for child in children:
                        if isinstance(child, NavigableString):
                            self.entry.append(str(child))
                        elif isinstance(child, Tag) and child.name == "a":
                            self.entry.append(self.Hyperlink(child.text, child.attrs["href"]))
                        elif isinstance(child, Tag) and child.name == "span":
                            _children = list(child.children)
                            self.entry.append(self.Reference(int(_children[0].text[1:-1]), _children[1].text))
                elif tag.name == "img":
                    self.entry.append(self.Image(tag.attrs["src"], tag.attrs["title"]))

        self.title = soup.find("h2", {"id": "title"}).a.text
        self.question = soup.find("p", {"id": "question"}).text
        self.author = soup.find("p", {"id": "attribute"}).text[1:]
        self.url = f"{WHAT_IF_BASE_URL}{self.number}"
